#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <lv_log.h>
#include <lv_list.h>
#include <lv_audio.h>
#include <lv_video.h>
#include <lv_palette.h>
#include <lv_plugin.h>
#include <lv_songinfo.h>
#include <lv_actor.h>

extern VisList *__lv_plugins_actor;

static VisActorPlugin *get_actor_plugin (VisActor *actor);

static VisActorPlugin *get_actor_plugin (VisActor *actor)
{
	VisActorPlugin *actplugin;

	actplugin = actor->plugin->plugin.actorplugin;

	return actplugin;
}

VisList *visual_actor_get_list ()
{
	return __lv_plugins_actor;
}

char *visual_actor_get_next_by_name (char *name)
{
	return visual_plugin_get_next_by_name (visual_actor_get_list (), name);
}

char *visual_actor_get_prev_by_name (char *name)
{
	return visual_plugin_get_prev_by_name (visual_actor_get_list (), name);
}

int visual_actor_valid_by_name (char *name)
{
	if (_lv_plugin_find (visual_actor_get_list (), name) == NULL)
		return FALSE;
	else
		return TRUE;
}

VisActor *visual_actor_new (char *actorname)
{
	VisActor *actor;
	VisPluginRef *ref;

	if (__lv_plugins_actor == NULL && actorname != NULL) {
		visual_log (VISUAL_LOG_LEVEL_CRITICAL, "the plugin list is NULL");
		return NULL;
	}
	
	actor = malloc (sizeof (VisActor));
	memset (actor, 0, sizeof (VisActor));
	
	if (actorname == NULL)
		return actor;

	ref = _lv_plugin_find (__lv_plugins_actor, actorname);

	actor->plugin = _lv_plugin_load (ref);

	return actor;
}

int visual_actor_realize (VisActor *actor)
{
	if (actor == NULL) {
		visual_log (VISUAL_LOG_LEVEL_CRITICAL, "the actor is NULL");
		return -1;
	}

	if (actor->plugin == NULL) {
		visual_log (VISUAL_LOG_LEVEL_CRITICAL, "the actor->plugin element is NULL");
		return -1;
	}

	_lv_plugin_realize (actor->plugin);

	return 0;
}

int visual_actor_destroy (VisActor *actor)
{
	if (actor == NULL)
		return -1;

	if (actor->plugin != NULL)
		_lv_plugin_unload (actor->plugin);

	return visual_actor_free (actor);
}

int visual_actor_free (VisActor *actor)
{
	if (actor == NULL)
		return -1;

	if (actor->transform != NULL)
		visual_video_free_with_buffer (actor->transform);

	if (actor->fitting != NULL)
		visual_video_free_with_buffer (actor->fitting);
		
	if (actor->songinfo != NULL)
		visual_songinfo_free (actor->songinfo);

	if (actor->songcompare != NULL)
		visual_songinfo_free (actor->songcompare);

	free (actor);
	
	return 0;
}

int visual_actor_events_pump (VisActor *actor)
{
	VisActorPlugin *actplugin;

	if (actor == NULL)
		return -1;

	actplugin = get_actor_plugin (actor);

	if (actplugin->events != NULL)
		actplugin->events (actplugin, &actor->events);

	return 0;
}

VisEventQueue *visual_actor_get_eventqueue (VisActor *actor)
{
	if (actor == NULL)
		return NULL;

	return &actor->events;
}

VisSongInfo *visual_actor_get_songinfo (VisActor *actor)
{
	if (actor == NULL)
		return NULL;

	if (actor->songinfo == NULL)
		actor->songinfo = visual_songinfo_new (VISUAL_SONGINFO_TYPE_NULL);
	
	return actor->songinfo;
}

VisPalette *visual_actor_get_palette (VisActor *actor)
{
	VisActorPlugin *actplugin;

	if (actor == NULL)
		return NULL;

	actplugin = get_actor_plugin (actor);
	
	if (actor->transform != NULL &&
		actor->video->depth == VISUAL_VIDEO_CONTEXT_8BIT) {
		
		return actor->ditherpal;

	} else {
		return actplugin->palette (actplugin);
	}

	return NULL;
}

int visual_actor_video_negotiate (VisActor *actor, int rundepth, int noevent, int forced)
{
	VisActorPlugin *actplugin = get_actor_plugin (actor);
	VisVideo tempv;
	int tmpwidth, tmpheight, tmppitch;
	int depthflag;

	if (actor->transform != NULL) {
		visual_video_free_with_buffer (actor->transform);
		actor->transform = NULL;
	}
	
	if (actor->fitting != NULL) {
		visual_video_free_with_buffer (actor->fitting);
		actor->fitting = NULL;
	}

	if (actor->ditherpal != NULL) {
		visual_palette_free (actor->ditherpal);
		actor->ditherpal = NULL;
	}

	depthflag = visual_actor_get_supported_depth (actor);

	printf ("[actor-negotiate] negotiating plugin %s\n", actor->plugin->ref->name);
	/* Set up depth transformation enviroment */
	if (visual_video_depth_is_supported (depthflag, actor->video->depth) != TRUE ||
			(forced == TRUE && actor->video->depth != rundepth)) {

		/* Depth transform enviroment, it automaticly
		 * fits size because it can use the pitch from
		 * the dest video context */
		actor->transform = visual_video_new ();

		if (forced == TRUE)
			visual_video_set_depth (actor->transform, rundepth);
		else
			visual_video_set_depth (actor->transform,
					visual_video_depth_get_highest_nogl (depthflag));

		printf ("[actor-negotiate] transpitch1 %d\n", actor->transform->pitch);
		/* If there is only GL (which gets returned by highest nogl if
		 * nothing else is there, stop here */
		if (actor->transform->depth == VISUAL_VIDEO_CONTEXT_GL)
			return -1;

		visual_video_set_dimension (actor->transform, actor->video->width, actor->video->height);
		printf ("[actor-negotiate] transpitch2 %d %d\n", actor->transform->width, actor->transform->pitch);
		
		actplugin->requisition (actplugin, &actor->transform->width, &actor->transform->height);
		printf ("[actor-negotiate] transpitch3 %d\n", actor->transform->pitch);
		
		if (noevent == FALSE) {
			visual_event_queue_add_resize (&actor->events, actor->transform,
					actor->transform->width, actor->transform->height);
			visual_actor_events_pump (actor);
		} else {
			/* Normally a visual_video_set_dimension get's called within the
			 * event handler, but we won't come there right now so we've
			 * got to set the pitch ourself */
			visual_video_set_dimension (actor->transform,
					actor->transform->width, actor->transform->height);
		}
		
		printf ("[actor-negotiate] rundepth: %d transpitch %d\n", rundepth, actor->transform->pitch);
		visual_video_allocate_buffer (actor->transform);

		if (actor->video->depth == VISUAL_VIDEO_CONTEXT_8BIT)
			actor->ditherpal = visual_palette_new ();

	} else {
		tmpwidth = actor->video->width;
		tmpheight = actor->video->height;
		tmppitch = actor->video->pitch;

		/* Pump the resize events and handle all the pending events */
		actplugin->requisition (actplugin, &actor->video->width, &actor->video->height);
		
		if (noevent == FALSE) {
			visual_event_queue_add_resize (&actor->events, actor->video,
					actor->video->width, actor->video->height);
			visual_actor_events_pump (actor);
		}

		/* Size fitting enviroment */
		if (tmpwidth != actor->video->width || tmpheight != actor->video->height) {
			actor->fitting = visual_video_new_with_buffer (actor->video->width,
					actor->video->height, actor->video->depth);

			visual_video_set_dimension (actor->video, tmpwidth, tmpheight);
		}

		/* Set the pitch seen this is the framebuffer context */
		visual_video_set_pitch (actor->video, tmppitch);
	}

	return 0;
}

int visual_actor_get_supported_depth (VisActor *actor)
{
	VisActorPlugin *actplugin;

	if (actor == NULL)
		return -1;

	actplugin = get_actor_plugin (actor);

	return actplugin->depth;
}

int visual_actor_set_video (VisActor *actor, VisVideo *video)
{
	if (actor == NULL)
		return -1;

	actor->video = video;

	return 0;
}

int visual_actor_run (VisActor *actor, VisAudio *audio)
{
	VisActorPlugin *actplugin;
	VisVideo *video;
	VisVideo *transform;
	VisVideo *fitting;

	/* We don't check for video, because we don't always need a video */
	if (actor == NULL || audio == NULL)
		return -1;

	actplugin = get_actor_plugin (actor);

	/* Songinfo handling */
	if (actor->songinfo != NULL) {
		if (actor->songcompare == NULL) {
			visual_songinfo_mark (actor->songinfo);
		
			if (actplugin->events != NULL)
				visual_event_queue_add_newsong (&actor->events, actor->songinfo);

			actor->songcompare = visual_songinfo_new (VISUAL_SONGINFO_TYPE_NULL);
		}

		if (visual_songinfo_compare (actor->songinfo, actor->songcompare) != 0) {
			visual_songinfo_mark (actor->songinfo);

			if (actplugin->events != NULL)
				visual_event_queue_add_newsong (&actor->events, actor->songinfo);
		}
		/* FIXME should be done more effecient */
		visual_songinfo_free_strings (actor->songcompare);
		visual_songinfo_copy (actor->songcompare, actor->songinfo);
	}
	
	actplugin->songinfo = actor->songinfo;

	video = actor->video;
	transform = actor->transform;
	fitting = actor->fitting;

	visual_video_set_palette (video, visual_actor_get_palette (actor));
	
	visual_actor_events_pump (actor);

	/* Yeah some transformation magic is going on here when needed */
	if (transform != NULL && (transform->depth != video->depth)) {
		actplugin->render (actplugin, transform, audio);

		if (transform->depth == VISUAL_VIDEO_CONTEXT_8BIT) {
			visual_video_set_palette (transform, visual_actor_get_palette (actor));
			visual_video_depth_transform (video, transform);
		} else {
			visual_video_set_palette (transform, actor->ditherpal);
			visual_video_depth_transform (video, transform);
		}
	} else {
		if (fitting != NULL && (fitting->width != video->width || fitting->height != video->height)) {
			actplugin->render (actplugin, fitting, audio);
			visual_video_fit_in_video (video, fitting);
		} else {
			actplugin->render (actplugin, video, audio);
		}
	}

	return 0;
}

