/*	List implementation from RCL.
 *	
 *	Copyright (C) 2002,2003,2004
 *				Dennis Smit <synap@nerds-incorporated.org>,
 *			  	Sepp Wijnands <mrrazz@nerds-incorporated.org>,
 *			   	Tom Wimmenhove <nohup@nerds-incorporated.org>
 *
 *	$Id: lv_list.c,v 1.4 2004/05/12 11:25:17 synap Exp $
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <fcntl.h>

#include <lv_list.h>

VisList *visual_list_new ()
{
	VisList *list;

	list = malloc (sizeof (VisList));
	if (list == NULL)
		return NULL;

	memset (list, 0, sizeof (VisList));

	return list;
}

int visual_list_free (VisList *list)
{
	if (list == NULL)
		return -1;

	free (list);

	return 0;
}

int visual_list_init (VisList *list)
{
	if (list == NULL)
		return -1;

	memset (list, 0, sizeof (VisList));

	return 0;
}

/* destroy a list and also its elements */
int visual_list_destroy (VisList *list, visual_list_destroy_func_t destroyer)
{
	VisListEntry *le = NULL;
	void *elem;

	if (list == NULL)
		return -1;
		
	/* walk through the given list and call the destroyer for it */
	while ((elem = visual_list_next (list, &le)) != NULL) {
		if (destroyer)
			destroyer (elem);
		visual_list_delete (list, &le);
	}

	visual_list_free (list);

	return 0;
}

int visual_list_duplicate_header (VisList *parent, VisList *child)
{
	child->head = parent->head;
	child->tail = parent->tail;
	child->count = parent->count;
	
	return 0;
}

void *visual_list_get (VisList *list, int index)
{
	VisListEntry *le = NULL;
	void *data = NULL;
	int i, lc;

	lc = visual_list_count (list);

	if (lc <= 0)
		return NULL;
	if (lc - 1 < index)
		return NULL;
	
	for (i = 0; i <= index; i++) {
		data = visual_list_next (list, &le);
		
		if (data == NULL)
			return NULL;
	}

	return data;
}

int visual_list_add_at_begin (VisList *list, void *data)
{
	VisListEntry *current, *next;

	if (list == NULL)
		return -1;

	/* allocate memory for new list entry */
	current = malloc (sizeof (VisListEntry));
	if (current == NULL)
		return -1;

	/* clear out memory */
	memset (current, 0, sizeof (VisListEntry));

	/* assign data element */
	current->data = data;

	if (list->head == NULL) {
		list->head = current;
		list->tail = current;
	} else {
		next = list->head;

		current->next = next;
		list->head = current;
	}

	/* done */
	list->count++;
}

int visual_list_add (VisList *list, void *data)
{
	VisListEntry *current, *prev;
	
	if (list == NULL)
		return -1;

	/* allocate memory for new list entry */
	current = malloc (sizeof (VisListEntry));
	if (current == NULL)
		return -1;

	/* clear out memory */
	memset (current, 0, sizeof (VisListEntry));

	/* assign data element */
	current->data = data;

	/* add list entry to list */
	/* is this the first entry for this list ? */
	if (list->head == NULL) {
		list->head = current;
		list->tail = current;
	} else {
		/* nope, add to tail of this list */
		prev = list->tail;

		/* exchange pointers */
		prev->next = current;
		current->prev = prev;
		
		/* point tail to new entry */
		list->tail = current;
	}

	/* done */
	list->count++;

	return 0;
}

/* insert entry after 'le' */
int visual_list_insert (VisList *list, VisListEntry **le, void *data)
{
	VisListEntry *prev, *next, *current;
	
	if (list == NULL || le == NULL || data == NULL)
		return -1;
	
	current = malloc (sizeof (VisListEntry));
	if (current == NULL)
		return -1;

	/* clear out memory */
	memset (current, 0, sizeof (VisListEntry));

	/* assign data element */
	current->data = data;

	/* add entry to list */
	if (list->head == NULL && *le == NULL) {
		/* first entry */
		list->head = current;
		list->tail = current;
	} else if (*le == NULL) {
		/* insert entry at first position */
		next = list->head;
		/* exchange pointers */
		current->next = next;
		next->prev = current;
		/* point head to current pointer */
		list->head = current;
	} else {
		/* insert entry at *le's position */
		prev = *le;
		next = prev->next;
		
		current->prev = prev;
		current->next = next;

		prev->next = current;
		if (next != NULL)
			next->prev = current;
		else
			list->tail = current;
	}

	/* hop to new entry */
	*le = current;
	
	/* done */
	list->count++;
	return 0;
}


/* remove current entry from list */
void visual_list_delete(VisList *list, VisListEntry **le)
{
	VisListEntry *prev, *current, *next;
	
	if (list == NULL)
		return;
	
	prev = current = next = NULL;

	/* valid list entry ? */
	if (*le == NULL)
		return; /* nope */

	/* point new to le's previous entry */
	current = *le;
	prev = current->prev;
	next = current->next;

	/* does it have a previous entry ? */
	if (prev != NULL) 
		prev->next = next;
	else
		list->head = next;
	
	if (next != NULL) /* it does have a next entry ? */
		next->prev = prev;
	else
		list->tail = prev;

	/* point current entry to previous one */
	*le = prev;

	/* free 'old' pointer */
	list->count--;
	free (current);
}

/* binary search */
VisListEntry *visual_list_find (VisList *list, void *ptr)
{
	VisListEntry *le;
	void *p;
	
	if (list == NULL || ptr == NULL)
		return NULL;
	
	/* walk through list searching for 'ptr' */
	le = NULL;
	while ((p = visual_list_next (list, &le)) != NULL) {
		if (p != ptr)
			continue;
		
		/* found */
		return le;
	}

	/* not found */
	return NULL;
}

/* exchange two data pointers */
void visual_list_exchange_item (VisList *list, VisListEntry **a, VisListEntry **b)
{	void *data;
	
	
	if (list == NULL || a == NULL || b == NULL)
		return;

	/* save data from a */
	data = (*a)->data;

	/* exchange a with b */
	(*a)->data = (*b)->data;
	(*b)->data = data;

	/* done */
}

/* get list entry by item id and place it in le */
void visual_list_find_item_id (VisList *list, VisListEntry **le, int id)
{
	int count = 0;
	
	if (list == NULL || le == NULL)
		return;
	
	*le = NULL;
	while (visual_list_next(list, le) != NULL) {
		if (count == id)
			break;
		
		count++;
	}
}

/* return number of list entries */
int visual_list_count (VisList *list)
{
	VisListEntry *le = NULL;
	int count = 0;
	
	if (list == NULL)
		return -1;
	
	/* walk through list */
	while (visual_list_next (list, &le) != NULL) 
		count++;

	list->count = count;
	return count;
}

/* sort list's NOT REENTRANT - FIXME FIXME FIXME */
static visual_list_sort_func_t visual_list_sort_func;

int visual_list_sort_compare (const void *a, const void *b)
{
	VisListEntry *la = (VisListEntry *) *(void **)a;
	VisListEntry *lb = (VisListEntry *) *(void **)b;
	
	/* pass data from list entries to sorting function's */
	return visual_list_sort_func (la->data, lb->data); 
}

int visual_list_sort(VisList *list, visual_list_sort_func_t compare)
{
	VisListEntry *le, **al;
	int count, i, ptr = 0;
	
	if (list == NULL || compare == NULL)
		return -1;
	
	count = visual_list_count (list);
	if (count < 1)
		return -1;
	
	al = malloc (count * sizeof (VisListEntry *));
	if (al == NULL)
		return -1;

	/* clear pointer memory */
	memset (al, 0, count * sizeof (VisListEntry *));
	
	/* set sorting function to 'compare' */
	visual_list_sort_func = compare;
	
	/* copy all list entry pointers into al */
	le = list->head;
	while (le != NULL)  {
		al[ptr++] = le;
		le = le->next;
	}
	
	/* use qsort to sort this list for us */
	qsort (al, count, sizeof (VisListEntry *), visual_list_sort_compare);

	/* empty list, reconstruct it */
	for (i = 0; i < count - 1; i++) {
		al[i]->next = al[i+1];
		al[i+1]->prev = al[i];
	}
	
	al[0]->prev = NULL;
	al[count-1]->next = NULL;

	list->head = al[0];
	list->tail = al[count-1];

	/* done */
	free (al);
	return 0;
}

