#ifndef _LV_PLUGIN_H
#define _LV_PLUGIN_H

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <libvisual/lv_video.h>
#include <libvisual/lv_audio.h>
#include <libvisual/lv_palette.h>
#include <libvisual/lv_list.h>
#include <libvisual/lv_songinfo.h>
#include <libvisual/lv_event.h>

enum {
	VISUAL_PLUGIN_TYPE_NULL,
	VISUAL_PLUGIN_TYPE_ACTOR,
	VISUAL_PLUGIN_TYPE_INPUT,
	VISUAL_PLUGIN_TYPE_MORPH
};

typedef struct _VisPluginInfo VisPluginInfo;
typedef struct _VisPluginRef VisPluginRef;

typedef struct _VisActorPlugin VisActorPlugin;
typedef struct _VisInputPlugin VisInputPlugin;
typedef struct _VisMorphPlugin VisMorphPlugin;

typedef struct _LVPlugin LVPlugin;

/* Actor plugin methods */
typedef int (*plugin_actor_init_func_t)(VisActorPlugin *);
typedef int (*plugin_actor_cleanup_func_t)(VisActorPlugin *);
typedef int (*plugin_actor_requisition_func_t)(VisActorPlugin *, int *, int *);
typedef int (*plugin_actor_events_func_t)(VisActorPlugin *, VisEventQueue *);
typedef VisPalette *(*plugin_actor_palette_func_t)(VisActorPlugin *);
typedef int (*plugin_actor_render_func_t)(VisActorPlugin *, VisVideo *, VisAudio *);

/* Input plugin methods */
typedef int (*plugin_input_init_func_t)(VisInputPlugin *);
typedef int (*plugin_input_cleanup_func_t)(VisInputPlugin *);
typedef int (*plugin_input_upload_func_t)(VisInputPlugin *, VisAudio *);

/* Morph plugin methods */
typedef int (*plugin_morph_init_func_t)(VisMorphPlugin *);
typedef int (*plugin_morph_cleanup_func_t)(VisMorphPlugin *);
typedef int (*plugin_morph_palette_func_t)(VisMorphPlugin *, float, VisAudio *, VisPalette *, VisVideo *, VisVideo *);
typedef int (*plugin_morph_apply_func_t)(VisMorphPlugin *, float, VisAudio *, VisVideo *, VisVideo *, VisVideo *);

/* Plugin methods */
typedef LVPlugin *(*plugin_load_func_t)(VisPluginRef *);

struct _VisPluginInfo {
	char		*name;
	char		*author;
	char		*version;
	char		*about;
	char		*help;
};

struct _VisPluginRef {
	char		*file;
	char		*name;
	int		 usecount;
	int		 type;
	VisPluginInfo	*info;
};

struct _VisActorPlugin {
	char				*file;
	char				*name;
	VisPluginRef			*ref;
	VisPluginInfo			*info;
	plugin_actor_init_func_t	 init;
	plugin_actor_cleanup_func_t	 cleanup;
	plugin_actor_requisition_func_t	 requisition;
	plugin_actor_events_func_t	 events;
	plugin_actor_palette_func_t	 palette;
	plugin_actor_render_func_t	 render;
	void				*handle;
	int				 depth;
	int				 schedflags;
	int				 plugflags;
	int				 plugtype;
	VisSongInfo			*songinfo;
	void				*private;
};

struct _VisInputPlugin {
	char				*file;
	char				*name;
	VisPluginRef			*ref;
	VisPluginInfo			*info;
	plugin_input_init_func_t	 init;
	plugin_input_cleanup_func_t	 cleanup;
	plugin_input_upload_func_t	 upload;
	void				*handle;
	int				 plugflags;
	int				 plugtype;
	void				*private;
};

struct _VisMorphPlugin {
	char				*file;
	char				*name;
	VisPluginRef			*ref;
	VisPluginInfo			*info;
	plugin_morph_init_func_t	 init;
	plugin_morph_cleanup_func_t	 cleanup;
	plugin_morph_palette_func_t	 palette;
	plugin_morph_apply_func_t	 apply;
	void				*handle;
	int				 depth;
	int				 plugflags;
	int				 plugtype;
	void				*private;
};

struct _LVPlugin {
	VisPluginRef	*ref;
	void		*handle;
	int		 type;
	int		 realized;

	union {
		VisActorPlugin *actorplugin;
		VisInputPlugin *inputplugin;
		VisMorphPlugin *morphplugin;
	} plugin;
};

/* prototypes */
VisPluginInfo *visual_plugin_info_new (char *name, char *author, char *version, char *about, char *help);
VisPluginInfo *visual_plugin_info_duplicate (VisPluginInfo *pluginfo);
int visual_plugin_info_free (VisPluginInfo *pluginfo);

VisPluginRef *visual_plugin_ref_new ();
int visual_plugin_ref_free (VisPluginRef *ref);
int visual_plugin_ref_destroy (VisPluginRef *ref);
int visual_plugin_ref_list_destroy (VisList *list);

VisActorPlugin *visual_plugin_actor_new ();
int visual_plugin_actor_free (VisActorPlugin *actorplugin);

VisInputPlugin *visual_plugin_input_new ();
int visual_plugin_input_free (VisInputPlugin *inputplugin);

VisMorphPlugin *visual_plugin_morph_new ();
int visual_plugin_morph_free (VisMorphPlugin *morphplugin);

LVPlugin *visual_plugin_new ();
int visual_plugin_free (LVPlugin *plugin);

VisList *visual_plugin_get_registry ();
VisList *visual_plugin_registry_filter (VisList *pluglist, int type);

char *visual_plugin_get_next_by_name (VisList *list, char *name);
char *visual_plugin_get_prev_by_name (VisList *list, char *name);

int _lv_plugin_unload (LVPlugin *plugin);
LVPlugin *_lv_plugin_load (VisPluginRef *ref);
int _lv_plugin_realize (LVPlugin *plugin);
VisPluginRef *_lv_plugin_get_reference (VisPluginRef *refn, char *pluginpath);
VisList *_lv_plugin_get_list (char **paths, int pcount);
VisPluginRef *_lv_plugin_find (VisList *list, char *name);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _LV_PLUGIN_H */
