#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <lv_common.h>
#include <lv_video.h>

typedef struct {
	unsigned short b:5, g:6, r:5;
} _color16;

/* depth conversions */
static int depth_transform_8_to_16_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_8_to_24_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_8_to_32_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);

static int depth_transform_16_to_8_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_16_to_24_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_16_to_32_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);

static int depth_transform_24_to_8_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_24_to_16_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_24_to_32_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);

static int depth_transform_32_to_8_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_32_to_16_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);
static int depth_transform_32_to_24_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal);

VisVideo *visual_video_new ()
{
	VisVideo *video;

	video = malloc (sizeof (VisVideo));
	memset (video, 0, sizeof (VisVideo));

	return video;
}

VisVideo *visual_video_new_with_buffer (int width, int height, int depth)
{
	VisVideo *video;
	
	video = visual_video_new ();

	visual_video_set_depth (video, depth);
	visual_video_set_dimension (video, width, height);

	visual_video_allocate_buffer (video);

	return video;
}

int visual_video_free (VisVideo *video)
{
	if (video == NULL)
		return -1;

	free (video);

	return 0;
}

int visual_video_free_with_buffer (VisVideo *video)
{
	if (video == NULL)
		return -1;

	visual_video_free_buffer (video);

	free (video);

	return 0;
}

int visual_video_free_buffer (VisVideo *video)
{
	if (video == NULL || video->screenbuffer == NULL)
		return -1;

	free (video->screenbuffer);

	return 0;
}

int visual_video_allocate_buffer (VisVideo *video)
{
	if (video == NULL)
		return -1;

	printf ("[video-allocate-buffer] Allocating buffer with size: %d, width height bpp pitch %d %d %d %d calc %d\n", video->size,
			video->width, video->height, video->bpp, video->pitch, video->pitch * video->height);
	video->screenbuffer = malloc (video->size);
	memset (video->screenbuffer, 0, video->size);

	return 0;
}

int visual_video_clone (VisVideo *dest, VisVideo *src)
{
	if (dest == NULL || src == NULL)
		return -1;

	visual_video_set_depth (dest, src->depth);
	visual_video_set_dimension (dest, src->width, src->height);
	visual_video_set_pitch (dest, src->pitch);

	return 0;
}

int visual_video_set_palette (VisVideo *video, VisPalette *pal)
{
	if (video == NULL)
		return -1;

	video->pal = pal;

	return 0;
}

int visual_video_set_buffer (VisVideo *video, void *buffer)
{
	if (video == NULL)
		return -1;

	video->screenbuffer = buffer;

	return 0;
}

int visual_video_set_dimension (VisVideo *video, int width, int height)
{
	if (video == NULL)
		return -1;

	video->width = width;
	video->height = height;

	video->pitch = video->width * video->bpp;
	video->size = video->pitch * video->height;
	
	return 0;
}

int visual_video_set_pitch (VisVideo *video, int pitch)
{
	if (video == NULL)
		return -1;

	if (video->bpp <= 0)
		return -1;

	video->pitch = pitch;
	video->size = video->pitch * video->height;

	return 0;
}

int visual_video_set_depth (VisVideo *video, int depth)
{
	if (video == NULL)
		return -1;

	video->depth = depth;
	video->bpp = visual_video_bpp_from_depth (video->depth);

	return 0;
}

int visual_video_depth_is_supported (int depthflag, int depth)
{
	if (visual_video_depth_is_sane (depth) == 0)
		return -1;

	if ((depth & depthflag) > 0)
		return 1;

	return 0;
}

int visual_video_depth_get_next (int depthflag, int depth)
{
	int i = depth;
	
        if (visual_video_depth_is_sane (depth) == 0)
		                return -1;

	if (i == VISUAL_VIDEO_CONTEXT_NONE) {
		i = VISUAL_VIDEO_CONTEXT_8BIT;

		if ((i & depthflag) > 0)
			return i;
	}

	while (i < VISUAL_VIDEO_CONTEXT_GL) {
		i *= 2;

		if ((i & depthflag) > 0)
			return i;
	}

	return depth;
}

int visual_video_depth_get_previous (int depthflag, int depth)
{
	int i = depth;

	if (visual_video_depth_is_sane (depth) == 0)
		return -1;

	if (i == VISUAL_VIDEO_CONTEXT_NONE)
		return VISUAL_VIDEO_CONTEXT_NONE;

	while (i > VISUAL_VIDEO_CONTEXT_NONE) {
		i >>= 1;

		if ((i & depthflag) > 0)
			return i;
	}

	return depth;
}

int visual_video_depth_get_lowest (int depthflag)
{
	return visual_video_depth_get_next (depthflag, VISUAL_VIDEO_CONTEXT_NONE);
}

int visual_video_depth_get_highest (int depthflag)
{
	int highest = VISUAL_VIDEO_CONTEXT_8BIT;
	int i = 0;

	while (highest != i) {
		highest = i;

		i = visual_video_depth_get_next (depthflag, i);
	}

	return highest;
}

int visual_video_depth_get_highest_nogl (int depthflag)
{
	int depth;

	depth = visual_video_depth_get_highest (depthflag);

	return visual_video_depth_get_previous (depthflag, depth);
}

int visual_video_depth_is_sane (int depth)
{
	int count = 0;
	int i = 1;

	if (depth == 0)
		return 1;

	if (depth >= VISUAL_VIDEO_CONTEXT_ENDLIST)
		return 0;
	
	while (i < VISUAL_VIDEO_CONTEXT_ENDLIST) {
		if ((i & depth) > 0)
			count++;

		if (count > 1)
			return 0;

		i <<= 1;
	}

	return 1;
}

int visual_video_depth_value_from_enum (int enumval)
{
	switch (enumval) {
		case VISUAL_VIDEO_CONTEXT_8BIT:
			return 8;

		case VISUAL_VIDEO_CONTEXT_16BIT:
			return 16;

		case VISUAL_VIDEO_CONTEXT_24BIT:
			return 24;

		case VISUAL_VIDEO_CONTEXT_32BIT:
			return 32;

		default:
			return -1;
	}

	return -1;
}

int visual_video_depth_enum_from_value (int depth)
{
	switch (depth) {
		case 8:
			return VISUAL_VIDEO_CONTEXT_8BIT;

		case 16:
			return VISUAL_VIDEO_CONTEXT_16BIT;

		case 24:
			return VISUAL_VIDEO_CONTEXT_24BIT;

		case 32:
			return VISUAL_VIDEO_CONTEXT_32BIT;

		default:
			return -1;

	}

	return -1;
}

int visual_video_bpp_from_depth (int enumval)
{
	switch (enumval) {
		case VISUAL_VIDEO_CONTEXT_8BIT:
			return 1;
		
		case VISUAL_VIDEO_CONTEXT_16BIT:
			return 2;

		case VISUAL_VIDEO_CONTEXT_24BIT:
			return 3;

		case VISUAL_VIDEO_CONTEXT_32BIT:
			return 4;

		case VISUAL_VIDEO_CONTEXT_GL:
			return 0;

		default:
			return -1;
	}

	return -1;
}

/* FIXME: more screwing up, just fix this */
int visual_video_fit_in_video (VisVideo *dest, VisVideo *src)
{
	unsigned char *destr, *srcr;
	int space = 0;
	int spare;
	int pitchadd;
	int i, j, dind = 0, sind = 0;

	spare = dest->width - src->width;

	destr = dest->screenbuffer;
	srcr = src->screenbuffer;

	/* Calculate the spacer */
	if (spare % 2 != 0) {
		if (spare > 1)
			space = spare;
		else
			space = 1;
	} else {
		space = spare / 1;
	}

	/* The iamge doesn't fit */
	if (dest->width < src->width || dest->height < src->height)
		return -1;

	/* Place the image */
	pitchadd = ((dest->pitch / src->bpp) - dest->width) * src->bpp;

	// FIXME use memcpy here */
	for (i = 0; i < src->height; i++) {
		for (j = 0; j < src->width * src->bpp; j++) {
			destr[dind++] = srcr[sind++];
		}

		dind += space * src->bpp;
		dind += pitchadd;
	}

	return 0;
}

int visual_video_depth_transform (VisVideo *viddest, VisVideo *vidsrc)
{
	return visual_video_depth_transform_to_buffer (viddest->screenbuffer,
			vidsrc, vidsrc->pal, viddest->depth, viddest->pitch);
}

int visual_video_depth_transform_to_buffer (unsigned char *dest, VisVideo *video,
		VisPalette *pal, int destdepth, int pitch)
{
	unsigned char *srcbuf = video->screenbuffer;
	int width = video->width;
	int height = video->height;

	/* destdepth is equal to sourcedepth case */
	if (video->depth == destdepth) {
		memcpy (dest, video->screenbuffer, video->width * video->height * video->bpp);

		return 0;
	}

	if (video->depth == VISUAL_VIDEO_CONTEXT_8BIT) {

		if (destdepth == VISUAL_VIDEO_CONTEXT_16BIT)
			return depth_transform_8_to_16_c (dest, srcbuf, width, height, pitch, pal);

		if (destdepth == VISUAL_VIDEO_CONTEXT_24BIT)
			return depth_transform_8_to_24_c (dest, srcbuf, width, height, pitch, pal);

		if (destdepth == VISUAL_VIDEO_CONTEXT_32BIT)
			return depth_transform_8_to_32_c (dest, srcbuf, width, height, pitch, pal);

	} else if (video->depth == VISUAL_VIDEO_CONTEXT_16BIT) {
		
		if (destdepth == VISUAL_VIDEO_CONTEXT_8BIT)
			return depth_transform_16_to_8_c (dest, srcbuf, width, height, pitch, pal);

		if (destdepth == VISUAL_VIDEO_CONTEXT_24BIT)
			return depth_transform_16_to_24_c (dest, srcbuf, width, height, pitch, NULL);

		if (destdepth == VISUAL_VIDEO_CONTEXT_32BIT)
			return depth_transform_16_to_32_c (dest, srcbuf, width, height, pitch, NULL);
	
	} else if (video->depth == VISUAL_VIDEO_CONTEXT_24BIT) {

		if (destdepth == VISUAL_VIDEO_CONTEXT_8BIT)
			return depth_transform_24_to_8_c (dest, srcbuf, width, height, pitch, pal);

		if (destdepth == VISUAL_VIDEO_CONTEXT_16BIT)
			return depth_transform_24_to_16_c (dest, srcbuf, width, height, pitch, NULL);

		if (destdepth == VISUAL_VIDEO_CONTEXT_32BIT)
			return depth_transform_24_to_32_c (dest, srcbuf, width, height, pitch, NULL);

	} else if (video->depth == VISUAL_VIDEO_CONTEXT_32BIT) {

		if (destdepth == VISUAL_VIDEO_CONTEXT_8BIT)
			return depth_transform_32_to_8_c (dest, srcbuf, width, height, pitch, pal);

		if (destdepth == VISUAL_VIDEO_CONTEXT_16BIT)
			return depth_transform_32_to_16_c (dest, srcbuf, width, height, pitch, NULL);

		if (destdepth == VISUAL_VIDEO_CONTEXT_24BIT)
			return depth_transform_32_to_24_c (dest, srcbuf, width, height, pitch, NULL);
	}

	return -1;
}

/* Depth transform C code */
/* FIXME TODO depths:	c	sse	mmx	altivec
 * 8 - 16		x
 * 8 - 24		x
 * 8 - 32		x
 * 16 - 8		x
 * 16 - 24		x
 * 16 - 32		x
 * 24 - 8		x
 * 24 - 16		x
 * 24 - 32		x
 * 32 - 8		x
 * 32 - 24		x
 * 32 - 16		x
 */

static int depth_transform_8_to_16_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	_color16 *destr = (_color16 *) dest;
	int pitchdiff = pitch - (width * 2) >> 1;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			destr[i].r = pal->r[src[j]] >> 3;
			destr[i].g = pal->g[src[j]] >> 2;
			destr[i].b = pal->b[src[j]] >> 3;
			i++;
			j++;
		}

		i += pitchdiff;
	}
	
	return 0;
}

static int depth_transform_8_to_24_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	int pitchdiff = pitch - (width * 3);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			dest[i++] = pal->r[src[j]];
			dest[i++] = pal->g[src[j]];
			dest[i++] = pal->b[src[j]];
			j++;
		}

		i += pitchdiff;
	}

	return 0;
}

static int depth_transform_8_to_32_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	unsigned int *destr = (unsigned int *) dest;
	unsigned int col;
	int pitchdiff = pitch - (width * 4) >> 2;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			col = 0;
			col += pal->r[src[j]] << 16;
			col += pal->g[src[j]] << 8;
			col += pal->b[src[j]];
			j++;

			destr[i++] = col;
		}

		i += pitchdiff;
	}

	return 0;
}

static int depth_transform_16_to_8_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	_color16 *srcr = (_color16 *) src;
	unsigned char r, g, b;
	unsigned char col;
	int pitchdiff = pitch - width;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			r = srcr[j].r << 3;
			g = srcr[j].g << 2;
			b = srcr[j].b << 3;
			j++;

			col  = (r >> 5);
			col |= (g >> 6) << 3;
			col |= (b >> 5) << 5;

			pal->r[col] = r;
			pal->g[col] = g;
			pal->b[col] = b;

			dest[i++] = col;
		}

		i += pitchdiff;	
	}

	return 0;
}

static int depth_transform_16_to_24_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	_color16 *srcr = (_color16 *) src;
	int pitchdiff = pitch - (width * 3);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			dest[j++] = srcr[i].r << 3;
			dest[j++] = srcr[i].g << 2;
			dest[j++] = srcr[i].b << 3;
			i++;	
		}

		j += pitchdiff;
	}

	return 0;
}

static int depth_transform_16_to_32_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i, j = 0;
	_color16 *srcr = (_color16 *) src;
	int pitchdiff = pitch - (width * 4);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			dest[j++] = srcr[i].r << 3;
			dest[j++] = srcr[i].g << 2;
			dest[j++] = srcr[i].b << 3;
			dest[j++] = 0;
			i++;
		}
	
		j += pitchdiff;
	}
	
	return 0;
}

static int depth_transform_24_to_8_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	unsigned char r, g, b;
	unsigned char col;
	int pitchdiff = pitch - width;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			r = src[j++];
			g = src[j++];
			b = src[j++];

			col  = (r >> 5);
			col |= (g >> 6) << 3;
			col |= (b >> 5) << 5;

			pal->r[col] = r;
			pal->g[col] = g;
			pal->b[col] = b;

			dest[i++] = col;
		}

		i += pitchdiff;	
	}

	return 0;
}

static int depth_transform_24_to_16_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	_color16 *destr = (_color16 *) dest;
	int pitchdiff = (pitch - (width * 2)) >> 1;
	
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			destr[i].r = src[j++] >> 3;
			destr[i].g = src[j++] >> 2;
			destr[i].b = src[j++] >> 3;
			i++;
		}

		i += pitchdiff;
	}
	
	return 0;
}

static int depth_transform_24_to_32_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	int pitchdiff = pitch - (width * 4);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			dest[j++] = src[i++];
			dest[j++] = src[i++];
			dest[j++] = src[i++];
			dest[j++] = 0;
		}

		j += pitchdiff;
	}
	
	return 0;
}

static int depth_transform_32_to_8_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	unsigned char r, g, b;
	unsigned char col;
	int pitchdiff = pitch - width;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			r = src[j++];
			g = src[j++];
			b = src[j++];
			j++;

			col  = (r >> 5);
			col |= (g >> 6) << 3;
			col |= (b >> 5) << 5;

			pal->r[col] = r;
			pal->g[col] = g;
			pal->b[col] = b;

			dest[i++] = col;
		}

		i += pitchdiff;	
	}
	
	return 0;
}

static int depth_transform_32_to_16_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	_color16 *destr = (_color16 *) dest;
	int pitchdiff = (pitch - (width * 2)) >> 1;
	
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			destr[i].r = src[j++] >> 3;
			destr[i].g = src[j++] >> 2;
			destr[i].b = src[j++] >> 3;
			j++;
			i++;
		}

		i += pitchdiff;
	}

	return 0;
}

static int depth_transform_32_to_24_c (unsigned char *dest, unsigned char *src, int width, int height, int pitch, VisPalette *pal)
{
	int x, y;
	int i = 0, j = 0;
	int pitchdiff = pitch - (width * 3);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			dest[i++] = src[j++];
			dest[i++] = src[j++];
			dest[i++] = src[j++];
			j++;
		}
		
		i += pitchdiff;
	}
	
	return 0;
}

