#include <unistd.h>
#include <stdlib.h>

#include <time.h>
#include <sys/time.h>

#include <SDL/SDL.h>

#include <libvisual.h>

SDL_Surface *screen;
SDL_Color colors[256];

void sdl_init (int bpp)
{
	if (SDL_Init (SDL_INIT_VIDEO) < 0)
	{
		fprintf (stderr, "Unable to init SDL VIDEO: %s\n", SDL_GetError ());
		exit (0);
	}

	screen = SDL_SetVideoMode (320, 200, bpp, SDL_RESIZABLE);
}

void sdl_draw (VisVideo *video, unsigned char *buf, int unit1)
{
	unsigned char *str = (unsigned char *) screen->pixels;
	
	memcpy (str, buf, video->width * video->height * unit1);

	SDL_UpdateRect (screen, 0, 0, 0, 0);
}

void sdl_set_pal (VisPalette *pal)
{
	int i;

	for (i = 0; i < 256; i ++) {
		colors[i].r = pal->r[i];
		colors[i].g = pal->g[i];
		colors[i].b = pal->b[i];
	}

	SDL_SetColors (screen, colors, 0, 256);
}

int main (int argc, char *argv[])
{
	float alpha_var = 0.5;
	int morphing = 0;
	float morph_speed = 0.01;
	int morph_to = 0;

	time_t begin = time (NULL), end;

	VisPluginRef *ref;
	VisList *list;	

	VisInput *input;
	VisList *inlist;
	
	VisActorPlugin *actor1, *actor2;
	VisVideo *video1, *video2;
	VisVideo *tmpvid1, *tmpvid2;

	VisPalette *pal;
	
	unsigned char *buf1, *buf2;
	unsigned char *tmp1, *tmp2;
	unsigned char *scrbuf;
	
	int freeze = 0;
	
	int enddepth = VISUAL_VIDEO_CONTEXT_32BIT, bpp = 4;
	
	int frames = 0;
	
	SDL_Event event;

	if (argc < 2) {
		printf ("Usage: %s vis1 vis2 blenddepth\n", argv[0]);
		exit (-1);
	}

	if (argc > 2) {
		enddepth = visual_video_depth_enum_from_value (atoi (argv[3]));
		bpp = visual_video_bpp_from_enum (enddepth);

		printf ("depthidd %d, bpp %d\n", enddepth, bpp);
	}
	
	inlist = visual_input_get_list (NULL);
	input = visual_input_new (inlist, "esd");
	visual_input_realize (input);

	list = _lv_plugin_actor_get_list ("/usr/lib/libvisual/actor");
	
	ref = _lv_plugin_actor_find (list, argv[1]);
	actor1 = _lv_plugin_actor_load (ref);
	_lv_plugin_actor_realize (actor1);

	ref = _lv_plugin_actor_find (list, argv[2]);
	actor2 = _lv_plugin_actor_load (ref);
	_lv_plugin_actor_realize (actor2);

	buf1 = malloc (320 * 200 * 4);
	tmp1 = malloc (320 * 200 * bpp);
	video1 = visual_video_new ();
	visual_video_set_opts (video1, "depth", VISUAL_VIDEO_CONTEXT_32BIT);
	visual_video_set_dimension (video1, 320, 200);
	visual_video_set_buffer (video1, buf1);

	tmpvid1 = visual_video_new ();
	visual_video_set_opts (tmpvid1, "depth", enddepth);
	visual_video_set_dimension (tmpvid1, 320, 200);
	visual_video_set_buffer (tmpvid1, tmp1);

	buf2 = malloc (320 * 200 * 4);
	tmp2 = malloc (320 * 200 * bpp);
	video2 = visual_video_new ();
	visual_video_set_opts (video2, "depth", VISUAL_VIDEO_CONTEXT_32BIT);
	visual_video_set_dimension (video2, 320, 200);
	visual_video_set_buffer (video2, buf2);

	tmpvid2 = visual_video_new ();
	visual_video_set_opts (tmpvid2, "depth", enddepth);
	visual_video_set_dimension (tmpvid2, 320, 200);
	visual_video_set_buffer (tmpvid2, tmp2);
				  
	sdl_init (bpp * 8);

	actor1->dimension (actor1, video1, 320, 200);
	actor2->dimension (actor2, video2, 320, 200);

	scrbuf = malloc (320 * 200 * bpp);
	memset (scrbuf, 0, 320 * 200 * bpp);

	pal = visual_palette_new ();

	while (1) {
		visual_input_run (input);

		if (freeze == 0) {
			actor1->render (actor1, video1, input->audio);
			actor2->render (actor2, video2, input->audio);
		
			if (morphing == 1) {
				if (morph_to < alpha_var)
					alpha_var -= morph_speed;
				else
					alpha_var += morph_speed;

				if (alpha_var <= 0) {
					alpha_var = 0;
					morphing = 0;
				} else if (alpha_var >= 1) {
					alpha_var = 1;
					morphing = 0;
				}
			}

			if (video1->depth != enddepth) {
				visual_video_depth_transform_to_buffer (tmp1, video1, pal, enddepth);
				visual_video_depth_transform_to_buffer (tmp2, video2, pal, enddepth);

				visual_video_alpha_blend_video (scrbuf, tmpvid1, tmpvid2, alpha_var);
			} else {
				visual_video_alpha_blend_video (scrbuf, video1, video2, alpha_var);
			}
			
			sdl_set_pal (pal);
			sdl_draw (tmpvid1, scrbuf, bpp);

			frames++;
		}
		
		usleep (5000);

		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					switch (event.key.keysym.sym)
					{
						case SDLK_f:
							freeze = 1 - freeze;
							break;

						case SDLK_z:
							alpha_var += 0.05;

							if (alpha_var > 1)
								alpha_var = 1.00;

							break;

						case SDLK_x:
							alpha_var -= 0.05;

							if (alpha_var < 0)
								alpha_var = 0.00;

							break;
						case SDLK_a:
							morph_speed += 0.01;
							break;

						case SDLK_s:
							morph_speed -= 0.01;
							break;

						case SDLK_v:
							morphing = 1;
							morph_to = 1 - morph_to;
							break;

						case SDLK_F11:
							SDL_WM_ToggleFullScreen (screen);
							break;

						case SDLK_p:
							end = time (NULL);

							printf ("Drawn %d frames in %d seconds, average fps %d\n",
									frames, end - begin, frames / (end - begin));
							break;

						case SDLK_ESCAPE:
							goto out;
							break;
					}
					break;

				case SDL_VIDEORESIZE:
					SDL_FreeSurface (screen);

					actor1->dimension (actor1, video1, event.resize.w, event.resize.h);
					actor2->dimension (actor2, video2, video1->width, video1->height);

					free (buf1);
					free (tmp1);
					free (buf2);
					free (tmp2);
					free (scrbuf);
					
					buf1 = malloc (video1->width * video1->height * 4);
					tmp1 = malloc (video1->width * video1->height * bpp);

					buf2 = malloc (video1->width * video1->height * 4);
					tmp2 = malloc (video1->width * video1->height * bpp);

					scrbuf = malloc (video1->width * video1->height * bpp);

					visual_video_set_buffer (video1, buf1);
					visual_video_set_buffer (video2, buf2);
						
					visual_video_set_buffer (tmpvid1, tmp1);
					visual_video_set_dimension (tmpvid1, video1->width, video1->height);
						
					visual_video_set_buffer (tmpvid2, tmp2);
					visual_video_set_dimension (tmpvid2, video1->width, video1->height);

					screen = SDL_SetVideoMode (video1->width, video1->height, bpp * 8, SDL_RESIZABLE);

					break;

				case SDL_QUIT:
					goto out;
					break;
			}
		}
	}
out:
	
	end = time (NULL);

	printf ("Drawn %d frames in %d seconds, average fps %d\n",
			frames, end - begin, frames / (end - begin));

	free (scrbuf);
	visual_video_free (video1);
	visual_video_free (video2);
	visual_video_free (tmpvid1);
	visual_video_free (tmpvid2);

	SDL_Quit ();

	visual_input_destroy (input);
	_lv_plugin_actor_unload (actor1);
	_lv_plugin_actor_unload (actor2);

	visual_input_list_destroy (inlist);
}

