#include <unistd.h>
#include <stdlib.h>

#include <time.h>
#include <sys/time.h>

#include <SDL/SDL.h>

#include <libvisual.h>

SDL_Surface *screen;
SDL_Color colors[256];

void sdl_init (void)
{
	if (SDL_Init (SDL_INIT_VIDEO) < 0)
	{
		fprintf (stderr, "Unable to init SDL VIDEO: %s\n", SDL_GetError ());
		exit (0);
	}

	screen = SDL_SetVideoMode (320, 200, 8, SDL_RESIZABLE);
}

void sdl_draw (VisVideo *video, unsigned char *buf)
{
	unsigned char *str = (unsigned char *) screen->pixels;
	
	memcpy (str, buf, video->width * video->height);

	SDL_UpdateRect (screen, 0, 0, 0, 0);
}

void sdl_set_pal (VisPalette *pal)
{
	int i;

	for (i = 0; i < 256; i ++) {
		colors[i].r = pal->r[i];
		colors[i].g = pal->g[i];
		colors[i].b = pal->b[i];
	}
	
	SDL_SetColors (screen, colors, 0, 256);
}

int main (int argc, char *argv[])
{
	VisPluginRef *ref;
	VisActorPlugin *actor;
	VisPluginInfo *info;
	VisVideo *video;
	VisPalette *pal;
	unsigned char *scrbuf;
	time_t begin = time (NULL), end;
	VisInput *input;
	VisList *inlist;

	VisList *list;
	int i;
	
	int frames = 0;
	
	SDL_Event event;
	
	scrbuf = malloc (320 * 200);
	memset (scrbuf, 0, 320 * 200);
	video = visual_video_new ();
	visual_video_set_opts (video, "depth", VISUAL_VIDEO_CONTEXT_8BIT);
	visual_video_set_dimension (video, 320, 200);
	visual_video_set_buffer (video, scrbuf);

	inlist = visual_input_get_list (NULL);
	input = visual_input_new (inlist, "esd");
	visual_input_realize (input);

	list = _lv_plugin_actor_get_list ("/usr/lib/libvisual/actor");
	
	if (argc > 1)
		ref = _lv_plugin_actor_find (list, argv[1]);
	else
		ref = _lv_plugin_actor_find (list, "oinksie");

	actor = _lv_plugin_actor_load (ref);
	_lv_plugin_actor_realize (actor);

	printf ("Actor %p\n", actor);
	if (actor == NULL)
		exit (-1);

	sdl_init ();

	actor->dimension (actor, video, 320, 200);
	while (1) {
		visual_input_run (input);

		actor->render (actor, video, input->audio);

		pal = actor->palette (actor);
		
		if (pal != NULL)
			sdl_set_pal (pal);
		
		sdl_draw (video, scrbuf);
		frames++;
		
		usleep (5000);

		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					switch (event.key.keysym.sym)
					{
						case SDLK_F11:
							SDL_WM_ToggleFullScreen (screen);
							break;
						
						case SDLK_p:
							end = time (NULL);

							printf ("Drawn %d frames in %d seconds, average fps %d\n",
									frames, end - begin, frames / (end - begin));
							break;

						case SDLK_ESCAPE:
							goto out;
							break;
					}
					break;

				case SDL_VIDEORESIZE:
					SDL_FreeSurface (screen);

					actor->dimension (actor, video, event.resize.w, event.resize.h);

					free (scrbuf);
					scrbuf = malloc (video->width * video->height);
					memset (scrbuf, 0, video->width * video->height);
					video->screenbuffer = scrbuf;

					screen = SDL_SetVideoMode (video->width, video->height, 8, SDL_RESIZABLE);
					sdl_draw (video, scrbuf);

					break;

				case SDL_QUIT:
					goto out;
					break;
			}
		}
	}

out:
	end = time (NULL);

	printf ("Drawn %d frames in %d seconds, average fps %d\n",
			frames, end - begin, frames / (end - begin));

	visual_video_free_with_buffer (video);

	SDL_Quit ();

	
	visual_input_destroy (input);
	_lv_plugin_actor_unload (actor);

	visual_input_list_destroy (inlist);
}

