#include <unistd.h>
#include <stdlib.h>

#include <time.h>
#include <sys/time.h>

#include <SDL/SDL.h>

#include <libvisual.h>

SDL_Surface *screen;
SDL_Color colors[256];

void sdl_init (int bpp)
{
	if (SDL_Init (SDL_INIT_VIDEO) < 0)
	{
		fprintf (stderr, "Unable to init SDL VIDEO: %s\n", SDL_GetError ());
		exit (0);
	}

	screen = SDL_SetVideoMode (320, 200, bpp * 8, SDL_RESIZABLE);
}

void sdl_set_pal (VisPalette *pal)
{
	int i;

	for (i = 0; i < 256; i ++) {
		colors[i].r = pal->r[i];
		colors[i].g = pal->g[i];
		colors[i].b = pal->b[i];
	}

	SDL_SetColors (screen, colors, 0, 256);
}

void sdl_draw (VisVideo *video, unsigned char *buf, int bpp)
{
	unsigned char *str = (unsigned char *) screen->pixels;
	
	memcpy (str, buf, video->width * video->height * bpp);

	SDL_UpdateRect (screen, 0, 0, 0, 0);
}

int main (int argc, char *argv[])
{
	VisPluginRef *ref1, *ref2;
	VisActorPlugin *actor1;
	VisVideo *video1;
	VisPalette *pal1;

	VisActorPlugin *actor2;
	VisVideo *video2;
	VisPalette *pal2;

	unsigned char *drawbuf, *scrbuf1, *scrbuf2, *orgbuf1, *orgbuf2, *bufptr;

	VisInput *input;
	VisList *inlist;
	
	VisList *list;
	int i;
	
	int draw_alone = 0;
	int draw_blended = 1;
	
	float alpha_var = 0.50;
	
	float morph_speed = 0.01;
	int morphing = 0;
	int morph_to = 0;
	
	int stop_update = 0;

	int enddepth = VISUAL_VIDEO_CONTEXT_24BIT;
	int bpp;
	
	time_t begin = time (NULL), end;
	int frames = 0;
	
	SDL_Event event;

	/* Load the esd plugin */
	inlist = visual_input_get_list (NULL);
	input = visual_input_new (inlist, "esd");
	visual_input_realize (input);

	/* Load the actors */
	list = _lv_plugin_actor_get_list ("/usr/lib/libvisual/actor");
	
	if (argc == 2) {
		ref1 = _lv_plugin_actor_find (list, argv[1]);
		ref2 = _lv_plugin_actor_find (list, argv[1]);
	} else if (argc > 2) {
		ref1 = _lv_plugin_actor_find (list, argv[1]);
		ref2 = _lv_plugin_actor_find (list, argv[2]);
	} else {
		ref1 = _lv_plugin_actor_find (list, "infinite");
		ref2 = _lv_plugin_actor_find (list, "infinite");
	}

	if (argc > 3)
		enddepth = visual_video_depth_enum_from_value (atoi (argv[3]));

	bpp = visual_video_bpp_from_enum (enddepth);

	orgbuf1 = malloc (320 * 200);
	orgbuf2 = malloc (320 * 200);
	
	scrbuf1 = malloc (320 * 200 * bpp);
	scrbuf2 = malloc (320 * 200 * bpp);
	drawbuf = malloc (320 * 200 * bpp);

	video1 = visual_video_new ();
	visual_video_set_opts (video1, "depth", VISUAL_VIDEO_CONTEXT_8BIT);
	visual_video_set_dimension (video1, 320, 200);
	visual_video_set_buffer (video1, orgbuf1);
	
	video2 = visual_video_new ();
	visual_video_set_opts (video2, "depth", VISUAL_VIDEO_CONTEXT_8BIT);
	visual_video_set_dimension (video2, 320, 200);
	visual_video_set_buffer (video2, orgbuf2);

	actor1 = _lv_plugin_actor_load (ref1);
	_lv_plugin_actor_realize (actor1);
	actor2 = _lv_plugin_actor_load (ref2);
	_lv_plugin_actor_realize (actor2);

	if (actor1 == NULL || actor2 == NULL) {
		printf ("Actor is NULL\n");
		exit (-1);
	}

	sdl_init (bpp);

	actor1->dimension (actor1, video1, 320, 200);
	actor2->dimension (actor2, video2, 320, 200);

	while (1) {
		visual_input_run (input);

		actor1->render (actor1, video1, input->audio);
		pal1 = actor1->palette (actor1);

		actor2->render (actor2, video2, input->audio);
		pal2 = actor2->palette (actor2);

		sdl_set_pal (pal1);
		
		visual_video_depth_transform_to_buffer (scrbuf1, video1, pal1,
				enddepth);

		visual_video_depth_transform_to_buffer (scrbuf2, video2, pal2,
				enddepth);

		if (morphing == 1) {
			if (morph_to < alpha_var)
				alpha_var -= morph_speed;
			else
				alpha_var += morph_speed;
		
			if (alpha_var <= 0) {
				alpha_var = 0;
				morphing = 0;
			} else if (alpha_var >= 1) {
				alpha_var = 1;
				morphing = 0;
			}
		}
		
		visual_video_alpha_blend_buffer (drawbuf, scrbuf1, scrbuf2, video1->size, enddepth, alpha_var);

		if (draw_blended == 1)
			bufptr = drawbuf;
		else if (draw_alone == 0)
			bufptr = scrbuf1;
		else
			bufptr = scrbuf2;

		if (stop_update == 0)
			sdl_draw (video1, bufptr, bpp);
		
		frames++;
		
		usleep (5000);

		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					switch (event.key.keysym.sym)
					{
						case SDLK_ESCAPE:
							goto out;
							break;

						case SDLK_F11:
							SDL_WM_ToggleFullScreen (screen);
							break;

						case SDLK_q:
							draw_alone = 0;
							break;

						case SDLK_w:
							draw_alone = 1;
							break;

						case SDLK_e:
							draw_blended = 1 - draw_blended;
							break;

						case SDLK_z:
							alpha_var += 0.05;

							if (alpha_var > 1)
								alpha_var = 1.00;

							break;

						case SDLK_x:
							alpha_var -= 0.05;

							if (alpha_var < 0)
								alpha_var = 0.00;

							break;
						case SDLK_a:
							morph_speed += 0.01;
							break;

						case SDLK_s:
							morph_speed -= 0.01;
							break;

						case SDLK_v:
							morphing = 1;
							morph_to = 1 - morph_to;
							break;

						case SDLK_f:
							stop_update = 1 - stop_update;
							break;

						case SDLK_p:
							end = time (NULL);

							printf ("Drawn %d frames in %d seconds, average fps %d\n",
									frames, end - begin, frames / (end - begin));
							break;
					}
					break;

				case SDL_VIDEORESIZE:
					SDL_FreeSurface (screen);

					actor1->dimension (actor1, video1, event.resize.w, event.resize.h);
					actor2->dimension (actor2, video2, event.resize.w, event.resize.h);

					free (orgbuf1);
					free (orgbuf2);
					
					free (scrbuf1);
					free (scrbuf2);
					free (drawbuf);

					orgbuf1 = malloc (video1->width * video1->height);
					orgbuf2 = malloc (video2->width * video2->height);

					scrbuf1 = malloc (video1->width * video1->height * bpp);
					scrbuf2 = malloc (video2->width * video2->height * bpp);
					drawbuf = malloc (video1->width * video1->height * bpp);
					
					video1->screenbuffer = orgbuf1;
					video2->screenbuffer = orgbuf2;

					screen = SDL_SetVideoMode (video1->width, video1->height, bpp * 8, SDL_RESIZABLE);

					break;

				case SDL_QUIT:
					goto out;
					break;
			}
		}
	}
out:
	end = time (NULL);

	printf ("Drawn %d frames in %d seconds, average fps %d\n",
			frames, end - begin, frames / (end - begin));

	visual_video_free_with_buffer (video1);
	visual_video_free_with_buffer (video2); 
	free (scrbuf1);
	free (scrbuf2);
	free (drawbuf);
	
	SDL_Quit ();

	visual_input_destroy (input);

	_lv_plugin_actor_unload (actor1);
	_lv_plugin_actor_unload (actor2);

	visual_input_list_destroy (inlist);
}

