#include <unistd.h>
#include <stdlib.h>

#include <time.h>
#include <sys/time.h>

#include <SDL/SDL.h>

#include <libvisual.h>

SDL_Surface *screen;
SDL_Color colors[256];

void sdl_init (void)
{
	if (SDL_Init (SDL_INIT_VIDEO) < 0)
	{
		fprintf (stderr, "Unable to init SDL VIDEO: %s\n", SDL_GetError ());
		exit (0);
	}

	screen = SDL_SetVideoMode (320, 200, 8, SDL_RESIZABLE);
}

void sdl_draw (VisVideo *video, unsigned char *buf)
{
	unsigned char *str = (unsigned char *) screen->pixels;
	
	memcpy (str, buf, video->width * video->height);

	SDL_UpdateRect (screen, 0, 0, 0, 0);
}

void sdl_set_pal (VisPalette *pal)
{
	int i;

	for (i = 0; i < 256; i ++) {
		colors[i].r = pal->r[i];
		colors[i].g = pal->g[i];
		colors[i].b = pal->b[i];
	}
	
	SDL_SetColors (screen, colors, 0, 256);
}

int main (int argc, char *argv[])
{
	VisPluginRef *ref1, *ref2;
	VisActorPlugin *actor1;
	VisVideo *video1;
	VisPalette *pal1;

	VisActorPlugin *actor2;
	VisVideo *video2;
	VisPalette *pal2;

	unsigned char *orgbuf;

	VisInput *input;
	VisList *inlist;
	
	VisList *list;
	int i;
	
	time_t begin = time (NULL), end;
	int frames = 0;
	
	SDL_Event event;

	/* Load the esd plugin */
	inlist = visual_input_get_list (NULL);
	input = visual_input_new (inlist, "esd");
	visual_input_realize (input); 
	/* Load the actors */
	orgbuf = malloc (320 * 200);
	
	video1 = visual_video_new ();
	visual_video_set_opts (video1, "depth", VISUAL_VIDEO_CONTEXT_8BIT);
	visual_video_set_dimension (video1, 320, 200);
	visual_video_set_buffer (video1, orgbuf);
	
	video2 = visual_video_new ();
	visual_video_set_opts (video2, "depth", VISUAL_VIDEO_CONTEXT_8BIT);
	visual_video_set_dimension (video2, 320, 200);
	visual_video_set_buffer (video2, orgbuf);

	list = _lv_plugin_actor_get_list ("/usr/lib/libvisual/actor");
	
	if (argc == 2) {
		ref1 = _lv_plugin_actor_find (list, argv[1]);
		ref2 = _lv_plugin_actor_find (list, argv[1]);
	} else if (argc > 2) {
		ref1 = _lv_plugin_actor_find (list, argv[1]);
		ref2 = _lv_plugin_actor_find (list, argv[2]);
	} else {
		ref1 = _lv_plugin_actor_find (list, "jess");
		ref2 = _lv_plugin_actor_find (list, "oinksie");
	}

	actor1 = _lv_plugin_actor_load (ref1);
	_lv_plugin_actor_realize (actor1);
	actor2 = _lv_plugin_actor_load (ref2);
	_lv_plugin_actor_realize (actor2);

	if (actor1 == NULL || actor2 == NULL) {
		printf ("Actor is NULL\n");
		exit (-1);
	}

	sdl_init ();

	actor1->dimension (actor1, video1, 320, 200);
	actor2->dimension (actor2, video2, 320, 200);

	while (1) {
		visual_input_run (input);

		actor1->render (actor1, video1, input->audio);
		pal1 = actor1->palette (actor1);

		actor2->render (actor2, video1, input->audio);
		pal2 = actor2->palette (actor2);

		sdl_set_pal (pal2);
		sdl_draw (video1, video1->screenbuffer);
		frames++;
		
		usleep (5000);

		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					switch (event.key.keysym.sym)
					{
						case SDLK_F11:
							SDL_WM_ToggleFullScreen (screen);
							break;

						case SDLK_p:
							end = time (NULL);

							printf ("Drawn %d frames in %d seconds, average fps %d\n",
									frames, end - begin, frames / (end - begin));
							break;
						
						case SDLK_ESCAPE:
							goto out;
							break;
					}
					break;

				case SDL_VIDEORESIZE:
					SDL_FreeSurface (screen);

					actor1->dimension (actor1, video1, event.resize.w, event.resize.h);
					actor2->dimension (actor2, video2, event.resize.w, event.resize.h);

					free (orgbuf);
					
					orgbuf = malloc (video1->width * video1->height);
					printf ("VIDEO1: width %d height %d\n", video1->width, video1->height);
					printf ("VIDEO2: width %d height %d\n", video2->width, video2->height);
					video1->screenbuffer = orgbuf;
					video2->screenbuffer = orgbuf;

					screen = SDL_SetVideoMode (video1->width, video1->height, 8, SDL_RESIZABLE);

					break;

				case SDL_QUIT:
					goto out;	
					break;
			}
		}
	}
out:
	end = time (NULL);

	printf ("Drawn %d frames in %d seconds, average fps %d\n",
			frames, end - begin, frames / (end - begin));

	SDL_Quit ();

	visual_video_free (video1);
	visual_video_free_with_buffer (video2);

	visual_input_destroy (input);
	_lv_plugin_actor_unload (actor1);
	_lv_plugin_actor_unload (actor2);

	visual_input_list_destroy (inlist);
}

